{
CD Player
---------
(c)2003 S.V.Bulba
http://bulba.at.kz/
vorobey@mail.khstu.ru

Description:
------------
Example of using MCI API from MMSystem.dll to manage CD-ROMs
- Initializes MCI device only before starting playing
- Automatically uninitializes MCI device when Stop button pressed,
  playing of playlist comes to the end or other CD drive should be used
- Can open all tracks from given drives, filtering not audio tracks
- Can open *.CDAs via Open Files standard dialog
- Supports time and position visualisation and seeking
}

unit MainWin;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, MCI, StdCtrls, ExtCtrls, MMSystem;

type
  TPLItem = class(TObject)
  public
    Drive:integer;
    Track:DWORD;
  end;
  TForm1 = class(TForm)
    Panel1: TPanel;
    SpeedButton1: TSpeedButton;
    Label1: TLabel;
    LBCDTrackNum: TLabel;
    PrevButton: TSpeedButton;
    NextButton: TSpeedButton;
    LBTrackLen: TLabel;
    ProgrBar: TScrollBar;
    LBCurTime: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    CDDNum: TLabel;
    PlayList: TListBox;
    AddCDButton: TSpeedButton;
    ClearButton: TSpeedButton;
    AddButton: TSpeedButton;
    OpenDialog1: TOpenDialog;
    StopButton: TSpeedButton;
    PauseButton: TSpeedButton;
    AboutButton: TSpeedButton;
    procedure PrevButtonClick(Sender: TObject);
    procedure PlayButtonClick(Sender: TObject);
    procedure PauseButtonClick(Sender: TObject);
    procedure NextButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure AddCDButtonClick(Sender: TObject);
    procedure AddButtonClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure AboutButtonClick(Sender: TObject);
    procedure PlayCurrent;
    procedure FreeAllCD;
    procedure ClearPlaylist;
    procedure TimePosTimerProc(Sender: TObject);
    procedure StartTimePosTimer;
    procedure StopTimePosTimer;
    procedure ProgrBarScroll(Sender: TObject; ScrollCode: TScrollCode;
                                    var ScrollPos: Integer);
    function PlayNext:boolean;
    procedure WM_NEXT(var Msg: TMessage);message MM_MCINOTIFY;
    procedure PlayListKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure PlaySelected;
    procedure PlayListDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
  CurrentPLItem:integer = -1;
  CurrentCDTrack:DWORD = 0;
  CurrentCDDevice:integer = -1;
  TimePosTimer:TTimer;
  TimerStarted:boolean = False;
  ScrollMoving:boolean;

implementation

uses SelectCDs;

{$R *.DFM}

function TimeSToStr(t:real):string;
var
 ms:integer;
begin
ms := round(t);
SetLength(Result,4);
Result[4] := char(ms mod 10 + 48);
ms := ms div 10;
Result[3] := char(ms mod 6 + 48);
ms := ms div 6;
Result[2] := ':';
Result[1] := char(ms mod 10 + 48);
ms := ms div 10;
if ms = 0 then exit;
Result := char(ms mod 6 + 48) + Result;
ms := ms div 6;
if ms = 0 then exit;
Result := IntToStr(ms) + ':' + Result
end;

procedure TForm1.PrevButtonClick(Sender: TObject);
begin
if CurrentPLItem > 0 then
 begin
  Dec(CurrentPLItem);
  PlayCurrent
 end
end;

procedure TForm1.PlayButtonClick(Sender: TObject);
begin
PlayCurrent
end;

procedure TForm1.PauseButtonClick(Sender: TObject);
begin
if Length(CDDrives) = 0 then exit;
if CurrentCDDevice = -1 then exit;
if CDIDs[CurrentCDDevice] = 0 then exit;
CDSwitchPause(CurrentCDDevice,Handle);
if not PlayingPaused then
 PauseButton.Caption := 'Paus'
else
 PauseButton.Caption := 'Resu'
end;

procedure TForm1.NextButtonClick(Sender: TObject);
begin
PlayNext
end;

procedure TForm1.StopButtonClick(Sender: TObject);
begin
FreeAllCD
end;

procedure TForm1.AddCDButtonClick(Sender: TObject);
var
 i,j,n:integer;
 PLItem:TPLItem;
 s:string;
 WasInit:boolean;
begin
if Form2.ShowModal <> mrOk then exit;
for i := 0 to Length(CDDrives) - 1 do
 if Form2.ListBox1.Selected[i] then
  begin
   WasInit := CDIDs[i] <> 0;
   InitCDDevice(i);
   try
    for j := 1 to CDGetNumberOfTracks(i) do
     if IsAudioTrack(i,j) then
      begin
       PLItem := TPLItem.Create;
       PLItem.Drive := i;
       PLItem.Track := j;
       s := CDDrives[i] + ':\Track';
       if j < 10 then s := s + '0';
       s := s + IntToStr(j) + '.cda';
       n := PlayList.Items.Add(s);
       PlayList.Items.Objects[n] := PLItem
      end
   finally
    if not WasInit then CloseCDDevice(i)
   end
  end
end;

procedure TForm1.AddButtonClick(Sender: TObject);
type
 TCDA = packed record
  rID:array[0..3] of char;
  rLen:DWORD;
  CDAID:array[0..3] of char;
  fID:array[0..3] of char;
  fLen:DWORD;
  Version,TrackNum:word;
 end;
var
 i,j,k:integer;
 PLItem:TPLItem;
 D:string;
 f:file of TCDA;
 CDA:TCDA;
begin
if Length(CDDrives) = 0 then exit;
if OpenDialog1.Execute then
 begin
  OpenDialog1.InitialDir := ExtractFileDir(OpenDialog1.FileName);
  for i := 0 to OpenDialog1.Files.Count - 1 do
   begin
    PLItem := TPLItem.Create;
    try
     D := UpperCase(ExtractFileDrive(OpenDialog1.Files.Strings[i]));
     if Length(D) = 0 then D := CDDrives[0];
     k := 0;
     for j := 0 to Length(CDDrives) - 1 do
      if CDDrives[j] = D[1] then
       begin
        k := j;
        break
       end;
     PLItem.Drive := k;
     AssignFile(f,OpenDialog1.Files.Strings[i]);
     Reset(f);
     Read(f,CDA);
     CloseFile(f);
     if (CDA.rID = 'RIFF') and
        (CDA.CDAID = 'CDDA') and
        (CDA.fID = 'fmt ') and
        (CDA.Version = 1) then
      PLItem.Track := CDA.TrackNum
     else
      begin
       D := UpperCase(OpenDialog1.Files.Strings[i]);
       j := Pos('.CDA',D);
       if j >= 3 then
        PLItem.Track := StrToInt(Copy(D,j - 2,2))
       else
        raise Exception.Create('Open CDA error')
      end;
     k := PlayList.Items.Add(OpenDialog1.Files.Strings[i]);
     PlayList.Items.Objects[k] := PLItem
    except
     PLItem.Free
    end;
   end;
 end
end;

procedure TForm1.ClearButtonClick(Sender: TObject);
begin
FreeAllCD;
ClearPlaylist
end;

procedure TForm1.AboutButtonClick(Sender: TObject);
begin
Application.MessageBox('CD Player'#13#10'Program by Sergey Bulba'#13#10 +
              'http://bulba.at.kz/'#13#10'vorobey@mail.khstu.ru'#13#10 +
              '(c)2003 S.V.Bulba','About program',MB_OK)
end;

procedure TForm1.PlayCurrent;
var
 MSF:packed record
  case boolean of
  True: (MSF:DWORD);
  False:(M,S,F:byte);
 end;
begin
StopTimePosTimer;
if PlayList.Items.Count = 0 then exit;
if CurrentPLItem = -1 then CurrentPLItem := 0;
if CurrentPLItem >= PlayList.Items.Count then exit;
if (CurrentCDDevice >= 0) and
   (CurrentCDDevice <> TPLItem(PlayList.Items.Objects[CurrentPLItem]).Drive) then
 CloseCDDevice(CurrentCDDevice);
CurrentCDDevice := TPLItem(PlayList.Items.Objects[CurrentPLItem]).Drive;
InitCDDevice(CurrentCDDevice);
CurrentCDTrack := TPLItem(PlayList.Items.Objects[CurrentPLItem]).Track;
LBCDTrackNum.Caption := IntToSTr(CurrentCDTrack);
CDDNum.Caption := CDDrives[CurrentCDDevice] + ':';
CDPlayTrack(0,CurrentCDTrack,Handle);
PauseButton.Caption := 'Paus';
PlayList.ItemIndex := CurrentPLItem;
MSF.MSF := CDGetTrackLength(0,CurrentCDTrack);
ProgrBar.Max := MSF.F + (MSF.S + MSF.M * 60) * 75;
LBTrackLen.Caption := TimeSToStr(MSF.F / 75 + MSF.S + MSF.M * 60);
StartTimePosTimer
end;

procedure TForm1.FreeAllCD;
var
 i:integer;
begin
StopTimePosTimer;
PauseButton.Caption := 'Paus';
for i := 0 to Length(CDDrives) - 1 do
 CloseCDDevice(i)
end;

procedure TForm1.ClearPlaylist;
var
 i:integer;
begin
CurrentPLItem := -1;
for i := 0 to PlayList.Items.Count - 1 do
 TPLItem(PlayList.Items.Objects[i]).Free;
PlayList.Clear
end;

procedure TForm1.TimePosTimerProc(Sender: TObject);
var
 MSF:packed record
  case boolean of
  True: (TMSF:DWORD);
  False:(T,M,S,F:byte);
 end;
begin
if CDIDs[0] = 0 then exit;
try
 MSF.TMSF := CDGetPosition(0)
except
 FreeAllCD
end;
LBCurTime.Caption := TimeSToStr(MSF.F / 75 + MSF.S + MSF.M * 60);
if not ScrollMoving then
 ProgrBar.Position := MSF.F + (MSF.S + MSF.M * 60) * 75
end;

procedure TForm1.StartTimePosTimer;
begin
if TimerStarted then exit;
ScrollMoving := False;
ProgrBar.Position := 0;
LBCurTime.Caption := '0:00';
TimePosTimer := TTimer.Create(Self);
TimePosTimer.Interval := 100;
TimePosTimer.OnTimer := TimePosTimerProc;
TimerStarted := True
end;

procedure TForm1.StopTimePosTimer;
begin
if not TimerStarted then exit;
TimePosTimer.Free;
TimerStarted := False
end;

procedure TForm1.ProgrBarScroll(Sender: TObject; ScrollCode: TScrollCode;
  var ScrollPos: Integer);
var
 i:DWORD;
 MSF:packed record
  case boolean of
  True: (MSF:DWORD);
  False:(M,S,F:byte);
 end;
begin
if CurrentCDDevice = -1 then exit;
if CDIDs[CurrentCDDevice] = 0 then exit;
ScrollMoving := ScrollCode = scTrack;
if PlayingPaused then exit;
if ScrollCode = scEndScroll then
 begin
  i := ScrollPos;
  MSF.F := i mod 75;
  i := i div 75;
  MSF.S := i mod 60;
  MSF.M := i div 60;
  CDSetPosition(CurrentCDDevice,CurrentCDTrack,MSF.MSF,Handle)
 end
end;

function TForm1.PlayNext;
begin
Result := CurrentPLItem < PlayList.Items.Count - 1;
if Result then
 begin
  Inc(CurrentPLItem);
  PlayCurrent
 end
end;

procedure TForm1.WM_NEXT;
begin
if CDIDs[0] = 0 then exit;
if Msg.LParam <> integer(CDIDs[0]) then exit;
if Msg.WParam = MCI_NOTIFY_SUCCESSFUL then
 if not PlayNext then
  begin
   FreeAllCD;
   ProgrBar.Position := ProgrBar.Max;
   LBCurTime.Caption := LBTrackLen.Caption
  end
end;

procedure TForm1.PlaySelected;
begin
if PlayList.ItemIndex >= 0 then
 begin
  CurrentPLItem := PlayList.ItemIndex;
  PlayCurrent
 end
end;

procedure TForm1.PlayListKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
case Key of
VK_RETURN: PlaySelected;
VK_DELETE:
 if PlayList.ItemIndex >= 0 then
  begin
   if PlayList.ItemIndex = CurrentPLItem then FreeAllCD;
   if PlayList.ItemIndex <= CurrentPLItem then Dec(CurrentPLItem);
   TPLItem(PlayList.Items.Objects[PlayList.ItemIndex]).Free;
   PlayList.Items.Delete(PlayList.ItemIndex)
  end
end 
end;

procedure TForm1.PlayListDblClick(Sender: TObject);
begin
PlaySelected
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
FileMode := 0;
if Length(CDDrives) > 0 then
 OpenDialog1.InitialDir := CDDrives[0] + ':\'
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
FreeAllCD;
ClearPlaylist
end;

end.
